#include <string.h>
#include <stdio.h>
#include <math.h>

#ifdef _WINDOWS
#ifndef WINDOWS
#define WINDOWS
#endif
#endif
#include "ac_plugin.h"

#define BASRELIEF_EPSILON 0.00001

//---------------------------------------------------------------------------------------
// DDH: The below function was provided by Lisa and Andy from the AC3D community.
//      It has been slightly modified from the original version provided.
/*---------------------------------------------------------------------------
   Get Image Data

   Parameters: ACObject *pObj - object to get texture data from
               BYTE **pData - pointer to receive texture data
               int *iWidth - width of image in pixels
               int *iHeight - height of image in pixels
               int *iDepth - depth of pixels in bytes

      Returns: int error code
---------------------------------------------------------------------------*/
static int ddhbasrelief_getimagedata(int idx, BYTE **pData, int *iWidth, int *iHeight, int *iDepth)
{
    ACImage *pImage;	// image

    // get the image
    pImage = texture_id_to_image( idx );
    if ( pImage == NULL ) {
        display_message("Could not get image from object.");
        return -2;
    }

    // get image data
    if ( ! ac_entity_get_int_value(pImage, "data", (int *)pData) ) {
        display_message("Could not get image data.");
        return -3;
    }

    // store image descriptor
    ac_image_get_dim(pImage, iWidth, iHeight, iDepth);

    // success
    return 0;
}
// DDH: End of imported function
//---------------------------------------------------------------------------------------
// Create a bas relief from a texture
static void ddhbasrelief_go( int argc, const char* argv[] )
{

   if( argc < 5 )
   {
      message_dialog( "basrelief : insufficient parameters" );
      return;
   }

   // Get params
   int idx = atoi( argv[1] );
   float dist = 0;
   float height = 0;
   if( sscanf( argv[2], "%g", &dist ) != 1 )
   {
      message_dialog( "basrelief : could not read maximum width/depth %s", argv[2] );
      return;
   }
   if( sscanf( argv[3], "%g", &height ) != 1 )
   {
      message_dialog( "basrelief : could not read height %s", argv[3] );
      return;
   }
   int maxsiz = atoi( argv[4] );

   if( dist <= 0 )
   {
      message_dialog( "basrelief : distance must be greater than zero" );
      return;
   }
   if( height <= 0 )
   {
      message_dialog( "basrelief : height must be greater than zero" );
      return;
   }

   // Get the texture data
   BYTE *imgdata = NULL;
   int imgwidth = 0;
   int imgheight = 0;
   int imgdepth = 0;
   if( ddhbasrelief_getimagedata( idx, &imgdata, &imgwidth, &imgheight, &imgdepth ) != 0 )
   {
      message_dialog( "basrelief : could not get image data from the specified texture." );
      return;
   }

   if( imgwidth < 1 || imgheight < 1 )
   {
      message_dialog( "basrelief : the width/height (%d/%d) of the image is invalid", imgwidth, imgheight );
      return;
   }

   display_message( "basrelief : creating %dx%d grid from texture.", imgwidth, imgheight );

   // Scale the image
   int imgsiz = imgwidth * imgheight;
   int imgwidthOrig = imgwidth;
   int imgheightOrig = imgheight;
   int imgsizOrig = imgsiz;
   double imgscalex = 1.0;
   double imgscalez = 1.0;
   if( imgsiz > maxsiz )
   {
      // Scale the width/height to the maximum
      imgwidth = sqrt(( double )maxsiz ) * ((( double )imgwidthOrig ) / sqrt(( double )imgsizOrig ));
      imgheight = sqrt(( double )maxsiz ) * ((( double )imgheightOrig ) / sqrt(( double )imgsizOrig ));
      imgsiz = imgwidth * imgheight;
      imgscalex = (( double )imgwidth ) / (( double )imgwidthOrig );
      imgscalez = (( double )imgheight ) / (( double )imgheightOrig );
      message_dialog( "The number of pixels in the image (%d) exceeds the maximum number of vertices allowed (%d).  The image data will be averaged down to %d x %d vertices.", imgsizOrig, maxsiz, imgwidth, imgheight );
   }

   // Convert the maximum width/depth to the distance between vertices
   dist = dist / (( float )( max( imgwidth, imgheight )));

   // Calculate 3d heightmap width/height
   double w3d = ( imgwidth - 1 ) * dist;
   double h3d = ( imgheight - 1 ) * dist;

   // Calculate the width/height offsets from the origin
   double woffset = w3d / 2.0;
   double hoffset = h3d / 2.0;

   // Convert the image data to a 2d array of color intensities.
   double **imgdata2d = new double*[imgwidthOrig];
   for( int ix = 0; ix < imgwidthOrig; ++ix )
   {
      imgdata2d[ix] = new double[imgheightOrig];
   }

   int ictr = 0;
   for( int iy = 0; iy < imgheightOrig; ++iy )
   {
      for( int ix = 0; ix < imgwidthOrig; ++ix )
      {
         switch( imgdepth )
         {
            case 1:
            case 2:
               imgdata2d[ix][iy] = ( double )( imgdata[ictr] );
               break;
            default:
               // Multiply each element by its wavelength in angstroms
               imgdata2d[ix][iy] = ((( double )( imgdata[ictr] )) * 0.299 +
                                    (( double )( imgdata[ictr+1] )) * 0.587 +
                                    (( double )( imgdata[ictr+2] )) * 0.114 );
               break;
         }
         ictr += imgdepth;
      }
   }

   // Convert the 2d image data to a byte array of intensities from 0-255
   double *b = new double[imgsiz];
   ictr = 0;
   bool scaledx = ( imgscalex - BASRELIEF_EPSILON < 1.0 );
   bool scaledz = ( imgscalez - BASRELIEF_EPSILON < 1.0 );
   int coordx;          // scaled X position
   int coordz;          // scaled Z position
   int psampx;          // number of horizontal pixels to sample based on scale
   int psampz;          // number of vertical pixels to sample based on scale
   psampx = ( int )( min( 1.0, (( double )imgwidthOrig  ) / (( double )imgwidth  )) + 0.5 );
   psampz = ( int )( min( 1.0, (( double )imgheightOrig ) / (( double )imgheight )) + 0.5 );
   int psampsiz = psampx * psampz;
   double *psamparr = new double[psampsiz];
   for( int iy = 0; iy < imgheight; ++iy )
   {
      if( scaledz )
      {
         coordz = ( min( imgheightOrig - 1, (( double )iy ) / (( double )imgheight ) * (( double )imgheightOrig )) + ( psampz / 2 ));
      }
      else
      {
         coordz = iy;
      }
      for( int ix = 0; ix < imgwidth; ++ix )
      {
         if( scaledx )
         {
            coordx = ( min( imgwidthOrig - 1, (( double )ix ) / (( double )imgwidth ) * (( double )imgwidthOrig )) + ( psampx / 2 ));
         }
         else
         {
            coordx = ix;
         }
         // Average the values in this area
         int isamps = 0;
         for( int ia = 0; ia < psampz; ++ia )
         {
            int cz = coordz - psampz / 2 + ia;
            if( cz > -1 && cz < imgheightOrig )
            {
               for( int ib = 0; ib < psampx; ++ib )
               {
                  int cx = coordx - psampx / 2 + ib;
                  if( cx > -1 && cx < imgwidthOrig )
                  {
                     psamparr[isamps] = imgdata2d[cx][cz];
                     ++isamps;
                  }
               }
            }
         }
         double davg = 0;
         for( int ia = 0; ia < isamps; ++ia )
         {
            davg += psamparr[ia];
         }
         davg /= isamps;
         b[ictr] = davg;
         ++ictr;
      }
   }
   delete psamparr;

   // Remove the 2d array of RGB intensities
   for( int ix = 0; ix < imgheight; ++ix )
   {
      delete imgdata2d[ix];
   }
   delete imgdata2d;

   // Create a vertex array
   // Create a texture coordinate array
   Vertex ***v = new Vertex**[imgwidth];
   Point2 **uv = new Point2*[imgwidth];
   for( int ix = 0; ix < imgwidth; ++ix )
   {
      v[ix] = new Vertex*[imgheight];
      uv[ix] = new Point2[imgheight];
      float x = (( float )ix ) / (( float )imgwidth );
      for( int iy = 0; iy < imgheight; ++iy )
      {
         v[ix][iy] = NULL;
         uv[ix][iy].x = x;
         uv[ix][iy].y = (( float )iy ) / (( float )imgheight );
      }
   }

   // Create a new object
   ACObject *ob = ac_new_object();

   // Set the vertex data
   double heightscale = height / 255.0;
   ictr = imgsiz - 1;
   float z;
   for( int iy = imgheight - 1; iy >= 0; --iy )
   {
      z = (( float )iy ) * (( float )dist ) - ( float )hoffset;
      for( int ix = imgwidth - 1; ix >= 0; --ix )
      {
         // Texture mapping (1x1 texture)
         v[ix][iy] = new_vertex_set( (( float )ix ) * (( float )dist ) - ( float )woffset, ( float )( b[ictr] * heightscale ), -z );
         object_add_vertex_head( ob, v[ix][iy] );
         --ictr;
      }
   }

   // Done with height data
   delete b;

   // Add vertices to surfaces
   for( int ix = 0; ix < imgwidth - 2; ++ix )
   {
      for( int iy = 0; iy < imgheight - 2; ++iy )
      {
         Surface *s = new_surface();
         object_add_surface_head( ob, s );
         surface_add_vertex_head( s, v[ix][iy+1], uv[ix][iy+1].x, uv[ix][iy+1].y );
         surface_add_vertex_head( s, v[ix+1][iy+1], uv[ix+1][iy+1].x, uv[ix+1][iy+1].y );
         surface_add_vertex_head( s, v[ix+1][iy], uv[ix+1][iy].x, uv[ix+1][iy].y );
         surface_add_vertex_head( s, v[ix][iy], uv[ix][iy].x, uv[ix][iy].y );
      }
   }

   // Delete vertex array
   // Delete texture coordinate array
   for( int ix = 0; ix < imgwidth; ++ix )
   {
      delete v[ix];
      delete uv[ix];
   }
   delete v;
   delete uv;

   // Add the object
   ac_object_set_crease_angle( ob, 180 );
   object_calc_normals_force( ob );
   add_undoable_world_add_object( "bas relief from texture", ob );

   // Finish
   find_bounding_box();
   selected_calc_normals();
   redraw_all();
   display_status();

}
//---------------------------------------------------------------------------------------
// Fill texture list
static void ddhbasrelief_showwin()
{

   int numtex = ac_get_num_textures();
   if( numtex < 1 )
   {
      message_dialog( "basrelief : no textures are loaded.  Load a texture in AC3D to use as a base image." );
      return;
   }

   tcl_command( "ddhbasrelief_dialog" );

   char *ch = NULL;
   for( int ix = 0; ix < numtex; ++ix )
   {
      ACImage *i = texture_id_to_image( ix );
      ac_entity_get_string_value(( ACEntity * )i, "name", &ch );
      tcl_command( ".ddhbasrelief.frmlist.listtex insert end \"%s\"", ch );
   }

}
//---------------------------------------------------------------------------------------
AC3D_PLUGIN_FUNC int AC3DPluginInit( AC3DPluginInitData *d )
{

   static bool firstTime = true;
   if( firstTime )
   {

      char *path = ( char * )_strdup( d->plugin_dir_path );
#ifdef WINDOWS
      ac_unbackslash_string( path );
#endif

      tcl_command( "source \"%s/basrelief.tcl\"", path );

      ac_add_command( "ddhbasrelief_showwin", ddhbasrelief_showwin );
      ac_add_tools_menu_item( "Bas relief from texture...", "ac3d ddhbasrelief_showwin",
                              "Create a bas relief grid from a texture" );
      ac_add_command_full( "ddhbasrelief_go", ( void * )ddhbasrelief_go,
                            3, "argv", "<texture index> <distance> <height>",
                           "Create a bas relief grid from a texture" );

      firstTime = false;

   }

   return 0;

}
//---------------------------------------------------------------------------------------
AC3D_PLUGIN_FUNC int AC3DPluginExit()
{
   return 0;
}
//---------------------------------------------------------------------------------------
AC3D_PLUGIN_FUNC char *AC3DPluginAbout()
{
   return( "Bas relief from texture - version 1.0 - Dennis Hawthorne" );
}
//---------------------------------------------------------------------------------------
AC3D_PLUGIN_FUNC char *AC3DPluginInfo()
{
   return( "Bas relief from texture - Create a bas relief on a new grid from a texture." );
}
//---------------------------------------------------------------------------------------

#undef BASRELIEF_EPSILON

